#!/bin/sh
usc-run votes --govtrack --congress=101 --session=1989 --chamber=senate $@
usc-run votes --govtrack --congress=101 --session=1990 $@
usc-run votes --govtrack --congress=102 --session=1991 $@
usc-run votes --govtrack --congress=102 --session=1992 $@
usc-run votes --govtrack --congress=103 --session=1993 $@
usc-run votes --govtrack --congress=103 --session=1994 $@
usc-run votes --govtrack --congress=104 --session=1995 $@
usc-run votes --govtrack --congress=104 --session=1996 $@
usc-run votes --govtrack --congress=105 --session=1997 $@
usc-run votes --govtrack --congress=105 --session=1998 $@
usc-run votes --govtrack --congress=106 --session=1999 $@
usc-run votes --govtrack --congress=106 --session=2000 $@
usc-run votes --govtrack --congress=107 --session=2001 $@
usc-run votes --govtrack --congress=107 --session=2002 $@
usc-run votes --govtrack --congress=108 --session=2003 $@
usc-run votes --govtrack --congress=108 --session=2004 $@
usc-run votes --govtrack --congress=109 --session=2005 $@
usc-run votes --govtrack --congress=109 --session=2006 $@
usc-run votes --govtrack --congress=110 --session=2007 $@
usc-run votes --govtrack --congress=110 --session=2008 $@
usc-run votes --govtrack --congress=111 --session=2009 $@
usc-run votes --govtrack --congress=111 --session=2010 $@
usc-run votes --govtrack --congress=112 --session=2011 $@
usc-run votes --govtrack --congress=112 --session=2012 $@
usc-run votes --govtrack --congress=113 --session=2013 $@
usc-run votes --govtrack --congress=113 --session=2014 $@
